@@ -11,26 +11,33 @@ import logging
 import os
 import re
 import shutil
+import subprocess
 import tempfile
+import threading
 
-import cliapp
+from . import releases
 
 logger = logging.getLogger(__name__)
 
-STABLE_CODENAME = 'bullseye'
 
-
-def run(*args, **kwargs):
+def run(process_args, **kwargs):
     """Run a command."""
-    def log_stdout(data):
-        """Handle stdout data from executed command."""
-        logger.debug('> %s', str(data.decode().strip()))
 
-    def log_stderr(data):
-        """Handle stderr data from executed command."""
-        logger.debug('2> %s', str(data.decode().strip()))
+    def _output_logger(file_handle, out_buffer, prompt):
+        """Read from stdout/stderr file handle and print to stdout."""
+        while (True):
+            read_bytes = file_handle.readline()
+            if not read_bytes:
+                break
 
-    logger.info('Executing command - %s %s', args, kwargs)
+            out_buffer.append(read_bytes)
+            logger.debug(f'{prompt}%s', read_bytes.decode().strip())
+
+        file_handle.close()
+
+    logger.info('Executing command - %s %s', process_args, dict(kwargs))
+    stdin_buffer = kwargs.pop('input', None)
+    check = kwargs.pop('check', True)
     environ = kwargs['env'] if 'env' in kwargs else os.environ.copy()
     environ['LC_ALL'] = 'C'
     environ['LANGUAGE'] = 'C'
@@ -38,9 +45,37 @@ def run(*args, **kwargs):
     environ['DEBIAN_FRONTEND'] = 'noninteractive'
     environ['DEBCONF_NONINTERACTIVE_SEEN'] = 'true'
     kwargs['env'] = environ
-    kwargs['stdout_callback'] = log_stdout
-    kwargs['stderr_callback'] = log_stderr
-    return cliapp.runcmd(*args, **kwargs)
+    kwargs['stdout'] = subprocess.PIPE
+    kwargs['stderr'] = subprocess.PIPE
+    kwargs['stdin'] = subprocess.PIPE if stdin_buffer else None
+    process = subprocess.Popen(process_args, **kwargs)
+
+    stdout_buffer = []
+    stdout_thread = threading.Thread(
+        target=_output_logger, args=[process.stdout, stdout_buffer, '> '])
+    stdout_thread.start()
+    stderr_buffer = []
+    stderr_thread = threading.Thread(
+        target=_output_logger, args=[process.stderr, stderr_buffer, '2> '])
+    stderr_thread.start()
+    if stdin_buffer:
+        process.stdin.write(stdin_buffer)
+        process.stdin.close()
+
+    stdout_thread.join()
+    stderr_thread.join()
+    return_code = process.wait()
+
+    stdout_buffer = b''.join(stdout_buffer) or None
+    stderr_buffer = b''.join(stderr_buffer) or None
+
+    if check and return_code:
+        logger.error('Command failed: %s\n%s\n%s', ' '.join(process_args),
+                     stdout_buffer, stderr_buffer)
+        raise subprocess.CalledProcessError(return_code, process_args,
+                                            stdout_buffer, stderr_buffer)
+
+    return stdout_buffer
 
 
 def run_in_chroot(state, *args, **kwargs):
@@ -195,13 +230,13 @@ def loopback_setup(state):
 def force_release_partition_loop(loop_device):
     """Force release a partition mapping on a loop device."""
     logger.info('Force releasing partition on loop device %s', loop_device)
-    run(['dmsetup', 'remove', loop_device], ignore_fail=True)
+    run(['dmsetup', 'remove', loop_device], check=False)
 
 
 def force_release_loop_device(loop_device):
     """Force release of a loop setup for entire device"""
     logger.info('Force releasing loop setup for device %s', loop_device)
-    run(['losetup', '-d', loop_device], ignore_fail=True)
+    run(['losetup', '-d', loop_device], check=False)
 
 
 def loopback_teardown(image_file):
@@ -252,20 +287,20 @@ def mount_filesystem(state,
     schedule_cleanup(state, unmount_filesystem, device, mount_point)
 
 
-def unmount_filesystem(device, mount_point, ignore_fail=False):
+def unmount_filesystem(device, mount_point, check=True):
     """Unmount a filesystem."""
     logger.info('Unmounting device %s from mount point %s', device,
                 mount_point)
-    run(['umount', mount_point], ignore_fail=ignore_fail)
+    run(['umount', mount_point], check=check)
 
 
 def process_cleanup(state):
     """Kill all processes using a given mount point."""
     mount_point = state['mount_point']
     logger.info('Killing all processes on the mount point %s', mount_point)
-    run(['fuser', '-mvk', mount_point], ignore_fail=True)
+    run(['fuser', '-mvk', mount_point], check=False)
     # XXX: Twice seems to work better?
-    run(['fuser', '-mvk', mount_point], ignore_fail=True)
+    run(['fuser', '-mvk', mount_point], check=False)
 
 
 def setup_extra_storage(state, file_system_type, size):
@@ -292,7 +327,7 @@ def cleanup_extra_storage(state, loop_device, extra_storage_file):
 
     _btrfs_rebalance(mount_point)
     run(['btrfs', 'balance', 'start', '-mconvert=dup', mount_point],
-        ignore_fail=True)
+        check=False)
 
     # Remove the extra storage device from btrfs filesystem
     run(['btrfs', 'device', 'remove', loop_device, mount_point])
@@ -306,9 +341,9 @@ def _btrfs_rebalance(mount_point):
     """Re-balance btrfs filesystem."""
     for usage in range(0, 100, 20):
         run(['btrfs', 'balance', 'start', f'-musage={usage}', mount_point],
-            ignore_fail=True)
+            check=False)
         run(['btrfs', 'balance', 'start', f'-dusage={usage}', mount_point],
-            ignore_fail=True)
+            check=False)
 
 
 def debootstrap(state, architecture, distribution, variant, components,
@@ -326,12 +361,11 @@ def debootstrap(state, architecture, distribution, variant, components,
             '--include=' + ','.join(packages), distribution, target, mirror
         ])
     except (Exception, KeyboardInterrupt):
+        run(['tail', os.path.join(target, 'debootstrap/debootstrap.log')])
         logger.info(
             'Unmounting filesystems that may have been left by debootstrap')
         for path in ('proc', 'sys', 'etc/machine-id'):
-            unmount_filesystem(None,
-                               os.path.join(target, path),
-                               ignore_fail=True)
+            unmount_filesystem(None, os.path.join(target, path), check=False)
         raise
 
     # During bootstrap, /etc/machine-id path might be bind mounted.
@@ -339,7 +373,7 @@ def debootstrap(state, architecture, distribution, variant, components,
                      unmount_filesystem,
                      None,
                      os.path.join(target, 'etc/machine-id'),
-                     ignore_fail=True)
+                     check=False)
 
 
 @contextlib.contextmanager
@@ -369,11 +403,11 @@ def install_from_backports(state, packages):
                         " ".join(packages))
             run_in_chroot(state, [
                 'apt-get', 'install', '-y', *[
-                    f'{package}/{STABLE_CODENAME}-backports'
+                    f'{package}/{releases.STABLE_CODENAME}-backports'
                     for package in packages
                 ]
             ])
-        except cliapp.AppException as ex:
+        except cliapp.AppException:
             logger.info(
                 'Packages not found in backports. Installing from stable.')
             run_in_chroot(state, ['apt-get', 'install', '-y', *packages])
@@ -497,8 +531,8 @@ deb http://security.debian.org/debian-security/ {distribution}-security {compone
 deb-src http://security.debian.org/debian-security/ {distribution}-security {components}
 '''
     backports_template = f'''
-deb http://deb.debian.org/debian {STABLE_CODENAME}-backports main
-deb-src http://deb.debian.org/debian {STABLE_CODENAME}-backports main
+deb http://deb.debian.org/debian {releases.STABLE_CODENAME}-backports main
+deb-src http://deb.debian.org/debian {releases.STABLE_CODENAME}-backports main
 '''
     file_path = path_in_mount(state, 'etc/apt/sources.list')
     with open(file_path, 'w') as file_handle:
@@ -529,8 +563,7 @@ def setup_flash_kernel(state, machine_name, kernel_options,
     if kernel_options:
         stdin = 'flash-kernel flash-kernel/linux_cmdline string {}'.format(
             kernel_options)
-        run_in_chroot(state, ['debconf-set-selections'],
-                      feed_stdin=stdin.encode())
+        run_in_chroot(state, ['debconf-set-selections'], input=stdin.encode())
 
     run_in_chroot(state, ['apt-get', 'install', '-y', 'flash-kernel'])
 
@@ -570,14 +603,18 @@ def fill_free_space_with_zeros(state):
     """
     zeros_path = path_in_mount(state, 'ZEROS')
     logger.info('Fill space with zeros on %s', state['mount_point'])
-    run(['dd', 'if=/dev/zero', 'of=' + zeros_path, 'bs=1M'], ignore_fail=True)
+    run(['dd', 'if=/dev/zero', 'of=' + zeros_path, 'bs=1M'], check=False)
     run(['rm', '-f', zeros_path])
 
 
 def compress(archive_file, image_file):
     """Compress an image using xz."""
     logger.info('Compressing file %s to %s', image_file, archive_file)
-    command = ['xz', '--no-warn', '--threads=0', '-9', '--force']
+    # Take upto 50% RAM when doing compression. Otherwise, number of threads
+    # will be reduced to accommodate.
+    command = [
+        'xz', '--no-warn', '--threads=0', '--memlimit=50%', '-9', '--force'
+    ]
     run(command + [image_file])
 
 
